"""Module to load and init namespace at package level."""
from .ncit import NCIt  # noqa: F401
from .mondo import Mondo  # noqa: F401
from .do import DO  # noqa: F401
from .oncotree import OncoTree  # noqa: F401
from .omim import OMIM  # noqa: F401
from .merge import Merge  # noqa: F401
